/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/gl-headers.h: Header which wraps the individual OpenGL
 * headers, defining compatibility constants if necessary.
 */

#ifndef SIL_SRC_SYSDEP_OPENGL_GL_HEADERS_H
#define SIL_SRC_SYSDEP_OPENGL_GL_HEADERS_H

/*************************************************************************/
/*************************************************************************/

#define GL_GLEXT_PROTOTYPES

#ifdef SIL_OPENGL_ES

# ifdef SIL_PLATFORM_ANDROID
/* Android uses dlsym(...,RTLD_DEFAULT) (not RTLD_NEXT, which doesn't work
 * because of the load order) to search for symbols, so we need to declare
 * the GL functions as non-exported to prevent the stub functions defined
 * in dyngl.c from being found in the search and causing infinite recursion.
 * Note that the default definition of GL_APICALL (KHRONOS_APICALL) already
 * includes a visibility attribute, so we have to rewrite the definition
 * from scratch.  Also note that __NDK_FPABI__ is only present through NDK
 * r12. */
#  ifdef __NDK_FPABI__
#   define GL_APICALL  __attribute__((visibility("hidden"))) __NDK_FPABI__
#  else
#   define GL_APICALL  __attribute__((visibility("hidden")))
#  endif
# endif

# if defined(SIL_PLATFORM_IOS)

/* The iOS SDK (at least through version 9.0) doesn't include GLES 3.1, so
 * we rename all the EXT_separate_shader_objects declarations to the proper
 * GLES 3.1 names. */
#  define glBindProgramPipelineEXT  glBindProgramPipeline
#  define glDeleteProgramPipelinesEXT  glDeleteProgramPipelines
#  define glGetProgramPipelineivEXT  glGetProgramPipelineiv
#  define glGetProgramPipelineInfoLogEXT  glGetProgramPipelineInfoLog
#  define glGenProgramPipelinesEXT  glGenProgramPipelines
#  define glProgramParameteriEXT  glProgramParameteri
#  define glProgramUniform1iEXT  glProgramUniform1i
#  define glProgramUniform1fEXT  glProgramUniform1f
#  define glProgramUniform2fvEXT  glProgramUniform2fv
#  define glProgramUniform3fvEXT  glProgramUniform3fv
#  define glProgramUniform4fvEXT  glProgramUniform4fv
#  define glProgramUniformMatrix4fvEXT  glProgramUniformMatrix4fv
#  define glUseProgramStagesEXT  glUseProgramStages
#  define glValidateProgramPipelineEXT  glValidateProgramPipeline

#  include <OpenGLES/ES3/gl.h>
/* We need Apple's glext.h for Apple-specific extensions. */
#  include <OpenGLES/ES3/glext.h>

#  undef glBindProgramPipelineEXT
#  undef glDeleteProgramPipelinesEXT
#  undef glGetProgramPipelineivEXT
#  undef glGetProgramPipelineInfoLogEXT
#  undef glGenProgramPipelinesEXT
#  undef glProgramParameteriEXT
#  undef glProgramUniform1iEXT
#  undef glProgramUniform1fEXT
#  undef glProgramUniform2fvEXT
#  undef glProgramUniform3fvEXT
#  undef glProgramUniform4fvEXT
#  undef glProgramUniformMatrix4fvEXT
#  undef glUseProgramStagesEXT
#  undef glValidateProgramPipelineEXT

#  define GL_APICALL GL_API
/* Some iOS headers blindly include <ES2/gl.h>, which causes breakage after
 * including <ES3/gl.h>. */
#  define __gl_es20_h_

# else  // not iOS
#  include <GLES3/gl31.h>
# endif

# include "external/opengl-headers/GLES2/gl2ext.h"

# define GLAPI GL_APICALL
# define GLAPIENTRY GL_APIENTRY

/* OpenGL ES 2.0 does away with the fixed function pipeline. */
# define SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS

/* Declare GLES 2.0-specific extension functions so dyngl doesn't spout
 * warnings. */
extern void glDiscardFramebufferEXT(GLenum, GLsizei, const GLenum *);

/* The direct state access functions aren't (yet?) in GLES, so we need to
 * declare them ourselves. */
GLAPI void GLAPIENTRY glBindTextureUnit(GLuint unit, GLuint texture);
GLAPI GLenum GLAPIENTRY glCheckNamedFramebufferStatus(GLuint framebuffer, GLenum target);
GLAPI void GLAPIENTRY glCompressedTextureSubImage2D(GLuint texture, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLsizei imageSize, const void *data);
GLAPI void GLAPIENTRY glCopyTextureSubImage2D(GLuint texture, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glCreateFramebuffers(GLsizei n, GLuint *framebuffers);
GLAPI void GLAPIENTRY glCreateProgramPipelines(GLsizei n, GLuint *pipelines);
GLAPI void GLAPIENTRY glCreateRenderbuffers(GLsizei n, GLuint *renderbuffers);
GLAPI void GLAPIENTRY glCreateTextures(GLenum target, GLsizei n, GLuint *textures);
GLAPI void GLAPIENTRY glGenerateTextureMipmap(GLuint texture);
GLAPI void GLAPIENTRY glGetNamedBufferSubData(GLuint buffer, GLintptr offset, GLsizeiptr size, void *data);
GLAPI void GLAPIENTRY glGetTextureImage(GLuint texture, GLint level, GLenum format, GLenum type, GLsizei bufSize, void *pixels);
GLAPI void GLAPIENTRY glInvalidateNamedFramebufferData(GLuint framebuffer, GLsizei numAttachments, const GLenum *attachments);
GLAPI void GLAPIENTRY glNamedFramebufferRenderbuffer(GLuint framebuffer, GLenum attachment, GLenum renderbuffertarget, GLuint renderbuffer);
GLAPI void GLAPIENTRY glNamedFramebufferTexture(GLuint framebuffer, GLenum attachment, GLuint texture, GLint level);
GLAPI void GLAPIENTRY glNamedRenderbufferStorage(GLuint renderbuffer, GLenum internalformat, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glTextureParameteri(GLuint texture, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTextureStorage2D(GLuint texture, GLsizei levels, GLenum internalformat, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glTextureSubImage2D(GLuint texture, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const void *pixels);

/* Debug message functions all have the KHR suffix. */
# define glDebugMessageCallback            glDebugMessageCallbackKHR
# define glDebugMessageControl             glDebugMessageControlKHR
# define glDebugMessageInsert              glDebugMessageInsertKHR
# define GL_DEBUG_OUTPUT                   GL_DEBUG_OUTPUT_KHR
# define GL_DEBUG_OUTPUT_SYNCHRONOUS       GL_DEBUG_OUTPUT_SYNCHRONOUS_KHR
# define GL_DEBUG_SOURCE_API               GL_DEBUG_SOURCE_API_KHR
# define GL_DEBUG_SOURCE_WINDOW_SYSTEM     GL_DEBUG_SOURCE_WINDOW_SYSTEM_KHR
# define GL_DEBUG_SOURCE_SHADER_COMPILER   GL_DEBUG_SOURCE_SHADER_COMPILER_KHR
# define GL_DEBUG_SOURCE_THIRD_PARTY       GL_DEBUG_SOURCE_THIRD_PARTY_KHR
# define GL_DEBUG_SOURCE_APPLICATION       GL_DEBUG_SOURCE_APPLICATION_KHR
# define GL_DEBUG_SOURCE_OTHER             GL_DEBUG_SOURCE_OTHER_KHR
# define GL_DEBUG_TYPE_ERROR               GL_DEBUG_TYPE_ERROR_KHR
# define GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_KHR
# define GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR  GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_KHR
# define GL_DEBUG_TYPE_PORTABILITY         GL_DEBUG_TYPE_PORTABILITY_KHR
# define GL_DEBUG_TYPE_PERFORMANCE         GL_DEBUG_TYPE_PERFORMANCE_KHR
# define GL_DEBUG_TYPE_MARKER              GL_DEBUG_TYPE_MARKER_KHR
# define GL_DEBUG_TYPE_PUSH_GROUP          GL_DEBUG_TYPE_PUSH_GROUP_KHR
# define GL_DEBUG_TYPE_POP_GROUP           GL_DEBUG_TYPE_POP_GROUP_KHR
# define GL_DEBUG_TYPE_OTHER               GL_DEBUG_TYPE_OTHER_KHR
# define GL_DEBUG_SEVERITY_HIGH            GL_DEBUG_SEVERITY_HIGH_KHR
# define GL_DEBUG_SEVERITY_MEDIUM          GL_DEBUG_SEVERITY_MEDIUM_KHR
# define GL_DEBUG_SEVERITY_LOW             GL_DEBUG_SEVERITY_LOW_KHR
# define GL_DEBUG_SEVERITY_NOTIFICATION    GL_DEBUG_SEVERITY_NOTIFICATION_KHR

#else  // !SIL_OPENGL_ES

# if defined(SIL_PLATFORM_WINDOWS)
/* Native Windows <GL/gl.h> requires <windows.h> to be included beforehand. */
#  include "src/sysdep/windows/internal.h"
/* We look up all symbols dynamically and define the wrappers statically,
 * so drop the dllimport attribute. */
#  undef DECLSPEC_IMPORT
#  define DECLSPEC_IMPORT  /*nothing*/
# endif

# define __glext_h_  // Prevent SDK-provided <glext.h> from being included.
# if defined(SIL_PLATFORM_MACOSX)
#  include <OpenGL/gl3.h>
#  define SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
# else
/* Work around a typo in old Mesa gl.h (#if should be #ifdef). */
#  define GL_ARB_shader_objects 0
#  include <GL/gl.h>
#  undef GL_ARB_shader_objects
# endif
# undef __glext_h_
# include "external/opengl-headers/GL/glext.h"

# ifndef GLAPIENTRY
#  define GLAPIENTRY APIENTRY
# endif

# if defined(SIL_PLATFORM_WINDOWS)
#  undef GLAPIENTRY
#  ifdef __i386__
#   define GLAPIENTRY  __attribute__((__stdcall__))
#  else
#   define GLAPIENTRY  /*nothing*/
#  endif
#  undef DECLSPEC_IMPORT
#  define DECLSPEC_IMPORT  __declspec(dllimport)
# endif

/* Map glDepthRangef() to glDepthRange() for the benefit of old OpenGL
 * implementations that don't have the former. */
# define glDepthRangef  glDepthRange

#endif  // SIL_OPENGL_ES

/* Equivalent of GLAPI for local functions. */
#ifdef SIL_PLATFORM_ANDROID
# define GLAPI_STATIC  static __NDK_FPABI__
#else
# define GLAPI_STATIC  static
#endif

/* Define functions missing from system headers to invalid glEnable() calls
 * to (1) avoid compiler warnings and (2) trigger errors if the functions
 * are in fact called. */
#define _INVALID_GL_CALL(name,...) \
    (DLOG("Invalid GL call: %s", name), glEnable(GL_INVALID_ENUM))
#ifdef SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
# define glAlphaFunc(...)            (_INVALID_GL_CALL("glAlphaFunc"))
# define glColor4f(...)              (_INVALID_GL_CALL("glColor4f"))
# define glColorPointer(...)         (_INVALID_GL_CALL("glColorPointer"))
# define glDisableClientState(...)   (_INVALID_GL_CALL("glDisableClientState"))
# define glEnableClientState(...)    (_INVALID_GL_CALL("glEnableClientState"))
# define glFogf(...)                 (_INVALID_GL_CALL("glFogf"))
# define glFogfv(...)                (_INVALID_GL_CALL("glFogfv"))
# define glFogi(...)                 (_INVALID_GL_CALL("glFogi"))
# define glFogx(...)                 (_INVALID_GL_CALL("glFogx"))
# define glLoadIdentity(...)         (_INVALID_GL_CALL("glLoadIdentity"))
# define glLoadMatrixf(...)          (_INVALID_GL_CALL("glLoadMatrixf"))
# define glMatrixMode(...)           (_INVALID_GL_CALL("glMatrixMode"))
# define glMultMatrixf(...)          (_INVALID_GL_CALL("glMultMatrixf"))
# define glTexCoordPointer(...)      (_INVALID_GL_CALL("glTexCoordPointer"))
# define glTranslatef(...)           (_INVALID_GL_CALL("glTranslatef"))
# define glVertexPointer(...)        (_INVALID_GL_CALL("glVertexPointer"))
#endif
#ifdef SIL_OPENGL_ES
# define glGetBufferSubData(...)     (_INVALID_GL_CALL("glGetBufferSubData"))
# define glGetTexImage(...)          (_INVALID_GL_CALL("glGetTexImage"))
# define glPointSize(...)            (_INVALID_GL_CALL("glPointSize"))
#endif

/*-----------------------------------------------------------------------*/

/* Make sure optional constants are defined to _something_ to avoid
 * compilation errors.  (The constants aren't used if the respective
 * features are unavailable.)  We can't detect the presence or absence of
 * functions in this way, so system-specific code is responsible for
 * ensuring that all functions are available (even if no-ops) or defined
 * as macros. */

/* Constants for primitive type. */
#ifndef GL_QUAD_STRIP
# define GL_QUAD_STRIP  GL_INVALID_ENUM
#endif
#ifndef GL_QUADS
# define GL_QUADS  GL_INVALID_ENUM
#endif

/* Constants used by the fixed-function pipeline. */
#ifndef GL_ALIASED_LINE_SIZE_RANGE
# define GL_ALIASED_LINE_SIZE_RANGE  GL_INVALID_ENUM
#endif
#ifndef GL_ALIASED_POINT_SIZE_RANGE
# define GL_ALIASED_POINT_SIZE_RANGE  GL_INVALID_ENUM
#endif
#ifndef GL_ALPHA_TEST
# define GL_ALPHA_TEST  GL_INVALID_ENUM
#endif
#ifndef GL_COLOR_ARRAY
# define GL_COLOR_ARRAY  GL_INVALID_ENUM
#endif
#ifndef GL_FOG
# define GL_FOG  GL_INVALID_ENUM
#endif
#ifndef GL_FOG_MODE
# define GL_FOG_MODE  GL_INVALID_ENUM
#endif
#ifndef GL_MODELVIEW
# define GL_MODELVIEW  GL_INVALID_ENUM
#endif
#ifndef GL_PROJECTION
# define GL_PROJECTION  GL_INVALID_ENUM
#endif
#ifndef GL_TEXTURE_COORD_ARRAY
# define GL_TEXTURE_COORD_ARRAY  GL_INVALID_ENUM
#endif
#ifndef GL_VERTEX_ARRAY
# define GL_VERTEX_ARRAY  GL_INVALID_ENUM
#endif

/* Constants for pixel data types. */
#ifndef GL_ALPHA8
# define GL_ALPHA8  GL_INVALID_ENUM
#endif
#ifndef GL_BGRA
# ifdef GL_BGRA_EXT
#  define GL_BGRA  GL_BGRA_EXT
# else
#  define GL_BGRA  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_LUMINANCE
# define GL_LUMINANCE  GL_INVALID_ENUM
#endif
#ifndef GL_LUMINANCE8
# define GL_LUMINANCE8  GL_INVALID_ENUM
#endif
#ifndef GL_UNSIGNED_SHORT_5_6_5_REV
# ifdef GL_UNSIGNED_SHORT_5_6_5_REV_EXT
#  define GL_UNSIGNED_SHORT_5_6_5_REV  GL_UNSIGNED_SHORT_5_6_5_REV_EXT
# else
#  define GL_UNSIGNED_SHORT_5_6_5_REV  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_UNSIGNED_SHORT_1_5_5_5_REV
# ifdef GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT
#  define GL_UNSIGNED_SHORT_1_5_5_5_REV  GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT
# else
#  define GL_UNSIGNED_SHORT_1_5_5_5_REV  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_UNSIGNED_SHORT_4_4_4_4_REV
# ifdef GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT
#  define GL_UNSIGNED_SHORT_4_4_4_4_REV  GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT
# else
#  define GL_UNSIGNED_SHORT_4_4_4_4_REV  GL_INVALID_ENUM
# endif
#endif

/* Constants for texture formats.  Some of these are missing from system
 * headers even when they should be present, so insert the real values. */
#ifndef GL_COMPRESSED_RGB_S3TC_DXT1_EXT
# define GL_COMPRESSED_RGB_S3TC_DXT1_EXT  0x83F0
#endif
#ifndef GL_COMPRESSED_RGBA_S3TC_DXT3_EXT
# define GL_COMPRESSED_RGBA_S3TC_DXT3_EXT  0x83F2
#endif
#ifndef GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
# define GL_COMPRESSED_RGBA_S3TC_DXT5_EXT  0x83F3
#endif
#ifndef GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG
# define GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG  0x8C01
#endif
#ifndef GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG
# define GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG  0x8C03
#endif
#ifndef GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG
# define GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG  0x8C00
#endif
#ifndef GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG
# define GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG  0x8C02
#endif

/* Constants related to vertex/index buffers. */
#ifndef GL_WRITE_ONLY
# if defined(GL_WRITE_ONLY_ARB)
#  define GL_WRITE_ONLY  GL_WRITE_ONLY_ARB
# elif defined(GL_WRITE_ONLY_OES)
#  define GL_WRITE_ONLY  GL_WRITE_ONLY_OES
# else
#  define GL_WRITE_ONLY  GL_INVALID_ENUM
# endif
#endif

/* Constants related to shaders. */
#ifndef GL_ACTIVE_PROGRAM
# ifdef GL_ACTIVE_PROGRAM_EXT
#  define GL_ACTIVE_PROGRAM  GL_ACTIVE_PROGRAM_EXT
# else
#  define GL_ACTIVE_PROGRAM  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_ALL_SHADER_BITS
# ifdef GL_ALL_SHADER_BITS_EXT
#  define GL_ALL_SHADER_BITS  GL_ALL_SHADER_BITS_EXT
# else
#  define GL_ALL_SHADER_BITS  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_FRAGMENT_SHADER_BIT
# ifdef GL_FRAGMENT_SHADER_BIT_EXT
#  define GL_FRAGMENT_SHADER_BIT  GL_FRAGMENT_SHADER_BIT_EXT
# else
#  define GL_FRAGMENT_SHADER_BIT  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_PROGRAM_PIPELINE_BINDING
# ifdef GL_PROGRAM_PIPELINE_BINDING_EXT
#  define GL_PROGRAM_PIPELINE_BINDING  GL_PROGRAM_PIPELINE_BINDING_EXT
# else
#  define GL_PROGRAM_PIPELINE_BINDING  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_PROGRAM_POINT_SIZE
# define GL_PROGRAM_POINT_SIZE  GL_INVALID_ENUM
#endif
#ifndef GL_PROGRAM_SEPARABLE
# ifdef GL_PROGRAM_SEPARABLE_EXT
#  define GL_PROGRAM_SEPARABLE  GL_PROGRAM_SEPARABLE_EXT
# else
#  define GL_PROGRAM_SEPARABLE  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_VERTEX_SHADER_BIT
# ifdef GL_VERTEX_SHADER_BIT_EXT
#  define GL_VERTEX_SHADER_BIT  GL_VERTEX_SHADER_BIT_EXT
# else
#  define GL_VERTEX_SHADER_BIT  GL_INVALID_ENUM
# endif
#endif

/*-----------------------------------------------------------------------*/

/* Disable glGetError() if requested by the system configuration. */

#ifdef SIL_OPENGL_DISABLE_GETERROR
# undef glGetError
# define glGetError() GL_NO_ERROR
#endif

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_OPENGL_GL_HEADERS_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
