/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/log.c: Log writing routines for Windows, using
 * Windows system calls instead of stdio to avoid the risk of MinGW/MSVC
 * stdio library collisions.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/sysdep.h"
#include "src/sysdep/windows/internal.h"
#include "src/userdata.h"

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void *sys_log_open(const char *name)
{
    const char *dir = userdata_get_data_path();
    if (!dir) {
        return NULL;
    }

    /* Get the directories created if necessary. */
    int id = userdata_save_data(name, "", 0);
    if (id) {
        userdata_wait(id);
        (void) userdata_get_result(id);
    }

    char pathbuf[1000];
    if (strformat_check(pathbuf, sizeof(pathbuf), "%s%s", dir, name)) {
        HANDLE fh = CreateFile(pathbuf, FILE_WRITE_DATA, 0, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
        if (fh != INVALID_HANDLE_VALUE) {
            return fh;
        }
    }

    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_log_write(void *logfile, const char *message, int len)
{
    HANDLE fh = logfile ? (HANDLE)logfile : GetStdHandle(STD_ERROR_HANDLE);
    WriteFile(fh, message, len+1, (DWORD[1]){}, NULL);
}

/*-----------------------------------------------------------------------*/

void sys_log_close(void *logfile)
{
    CloseHandle(logfile);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
